# -*- coding: utf-8 -*-

import os

from ....installer import Installer

from ..index_command import IndexCommand


class MakeRequirementsCommand(IndexCommand):
    """
    Renders a requirements.txt file for testing purposes.

    make:requirements
        { --no-dev : Do not write dev dependencies }
    """

    def handle(self):
        installer = Installer(self, self._repository)

        deps = self.poet.pip_dependencies

        if not self.option('no-dev'):
            deps += self.poet.pip_dev_dependencies

        self.line('')
        packages = installer.resolve(deps)

        requirements = os.path.join(self.poet.base_dir, 'requirements.txt')

        with open(requirements, 'w') as f:
            for package in packages:
                version = package['version']
                if isinstance(version, dict):
                    version = '-e {}@{}#egg={}'.format(version['git'], version['rev'], package['name'])
                else:
                    version = '{}=={}'.format(package['name'], version)

                f.write(version + '\n')

        self.line(' - Created <info>requirements.txt</> file')



# -*- coding: utf-8 -*-

from cleo import InputOption

from ...repositories import PyPiRepository

from .command import Command


class IndexCommand(Command):

    def __init__(self):
        super(IndexCommand, self).__init__()

        self._repository = PyPiRepository()

    def configure(self):
        super(IndexCommand, self).configure()

        # Adding --i|index option
        self.add_option(
            'index', 'i',
            InputOption.VALUE_REQUIRED,
            'The index to use'
        )

    def execute(self, i, o):
        super(IndexCommand, self).execute(i, o)

        index = self.option('index')

        if index:
            self._repository = PyPiRepository(index)



# -*- coding: utf-8 -*-

import subprocess

from jinja2 import Environment, PackageLoader

from .._compat import decode, PY3K


TEMPLATE_ENV = Environment(
    loader=PackageLoader('poet', 'templates'),
    autoescape=False,
    lstrip_blocks=True,
    trim_blocks=True
)

TEMPLATE_ENV.globals.update({
    'isinstance': isinstance,
    'list': list,
    'sorted': sorted,
    'repr': repr
})


def call(args):
    """
    Calls a command and return the output.
    
    :param args: The command args.
    :type args: list
    
    :rtype: str 
    """
    output = subprocess.check_output(args, stderr=subprocess.STDOUT)

    if PY3K:
        return decode(output)

    return output


def template(name):
    """
    Returns a template given a name.
    
    :param name: The name of the template.
    :type name: str
    
    :rtype: jinja2.Template 
    """
    if not name.endswith('.jinja2'):
        name += '.jinja2'

    return TEMPLATE_ENV.get_template(name)



# -*- coding: utf-8 -*-

from cleo import Application as BaseApplication

from .commands import (
    AboutCommand,
    CheckCommand,
    InitCommand,
    InstallCommand,
    LockCommand,
    PackageCommand,
    PublishCommand,
    RequireCommand,
    SearchCommand,
    UpdateCommand
)
from .commands.make import MakeSetupCommand, MakeRequirementsCommand


class Application(BaseApplication):
    """
    The console application that handles the commands.
    """

    def get_default_commands(self):
        default_commands = super(Application, self).get_default_commands()

        return default_commands + [
            AboutCommand(),
            CheckCommand(),
            InitCommand(),
            InstallCommand(),
            LockCommand(),
            MakeRequirementsCommand(),
            MakeSetupCommand(),
            PackageCommand(),
            PublishCommand(),
            RequireCommand(),
            SearchCommand(),
            UpdateCommand()
        ]


if __name__ == '__main__':
    Application().run()



# -*- coding: utf-8 -*-

from ...installer import Installer

from .index_command import IndexCommand


class LockCommand(IndexCommand):
    """
    Lock the dependencies specified in <comment>poetry.toml</comment>.

    lock
        {--f|force : Force locking}
        { --no-progress : Do not output download progress. }
    """

    def handle(self):
        if self.has_lock() and not self.option('force'):
            return

        installer = Installer(
            self, self._repository,
            with_progress=not self.option('no-progress')
        )

        installer.lock(self.poet)



# -*- coding: utf-8 -*-

import os
import shutil

from ..._compat import Path
from .command import Command


class PackageCommand(Command):
    """
    Builds the package.

    package
        { --no-universal : Do not build a universal package. }
        { --no-wheels : Build only the source package. }
        { --no-progress : Do not output download progress. }
    """

    def handle(self):
        poet = self.poet

        # Cleaning egg-info
        egg_info = os.path.join(poet.base_dir, '{}.egg-info'.format(poet.name))

        if os.path.exists(egg_info):
            shutil.rmtree(egg_info)

        self.line('')

        if self.option('no-progress'):
            self.line('Building <info>{}</> (<comment>{}</>)'.format(poet.name, poet.version))

            self._build(poet)
        else:
            with self.spin(
                'Building <info>{}</> (<comment>{}</>)'.format(poet.name, poet.version),
                'Built <info>{}</> (<comment>{}</>)'.format(poet.name, poet.version)
            ):
                self._build(poet)

        self.line('')

        dist = Path(self.poet_file).parent / 'dist'
        releases = dist.glob('{}-{}*'.format(self.poet.name, self.poet.normalized_version))
        for release in releases:
            self.line('  - Created <info>{}</>'.format(release.name))

        self.line('')

    def _build(self, poet):
        poet.build(
            universal=not self.option('no-universal'),
            no_wheels=self.option('no-wheels')
        )



# -*- coding: utf-8 -*-

import os

from twine.commands.upload import upload
from twine.commands.register import register
from requests.exceptions import HTTPError

from ...publisher import Publisher

from .command import Command


class PublishCommand(Command):
    """
    Publish the package to the remote repository.

    publish
        {--r|repository=pypi : The repository to register the package to}
    """

    def handle(self):
        # Checking if package exists
        package = os.path.join(self.poet.base_dir, 'dist', self.poet.archive)

        if not os.path.exists(package):
            self.info('No package found. Building it.')
            self.line('')
            self.call('package')
        else:
            self.info('Package found. Proceeding with publishing.')

        self.line('')
        self.line(
            'Publishing <info>{}</> (<comment>{}</>) to <fg=blue>{}</>'
            .format(self.poet.name, self.poet.version, self.option('repository'))
        )

        publisher = Publisher(
            self.output,
            self.option('repository')
        )

        self.line('')

        try:
            publisher.upload(self.poet)
        except HTTPError as e:
            if (e.response.status_code not in (403, 400)
                or e.response.status_code == 400 and 'was ever registered' not in e.response.text):
                raise

            # It may be the first time we publish the package
            # We'll try to register it and go from there
            try:
                publisher.register(self.poet)
            except HTTPError:
                raise

            publisher.upload(self.poet)



# -*- coding: utf-8 -*-

from ...installer import Installer

from .index_command import IndexCommand


class UpdateCommand(IndexCommand):
    """
    Update dependencies as according to the <comment>poetry.toml</> file.

    update
        { packages?* : The packages to update }
        { --f|features=* : Features to install }
        { --no-progress : Do not output download progress. }
    """

    def handle(self):
        if not self.has_lock():
            # We do not have a lock file
            # Assuming installation
            return self.call('install', [('--features', self.option('features'))])

        features = []
        for feature in self.option('features'):
            if ' ' in feature:
                features += [f.replace('_', '-').lower() for f in feature.split(' ')]
            else:
                features.append(feature.replace('_', '-').lower())

        installer = Installer(
            self, self._repository,
            with_progress=not self.option('no-progress')
        )

        installer.update(packages=self.argument('packages'), features=features)



# -*- coding: utf-8 -*-

from .package import PipDependency, Dependency
from .poet import Poet


class Lock(Poet):

    def is_lock(self):
        return True

    def load(self):
        root = self._config['root']
        self._name = root['name']
        self._version = root['version']

        packages = self._config['package']

        for package in packages:
            constraint = {
                'optional': package.get('optional', False),
                'python': package.get('python', ['*'])
            }
            version = package['version']
            if isinstance(version, dict):
                constraint.update(version)
            else:
                constraint['version'] = version

            dep = Dependency(
                package['name'],
                constraint,
                category=package['category']
            )
            pip_dep = PipDependency(
                package['name'],
                constraint,
                category=package['category'],
                checksum=package.get('checksum')
            )

            if package['category'] == 'dev':
                self._dev_dependencies.append(dep)
                self._pip_dev_dependencies.append(pip_dep)
            else:
                self._dependencies.append(dep)
                self._pip_dependencies.append(pip_dep)

        self._features = self._get_features()



# -*- coding: utf-8 -*-

import os
import re
import warnings

from setuptools.dist import Distribution
from setuptools.extension import Extension
from pip.commands.wheel import WheelCommand
from pip.status_codes import SUCCESS
from semantic_version import Spec, Version

from .._compat import Path, PY2, encode
from ..utils.helpers import template


class Builder(object):
    """
    Tool to transform a poet file to a setup() instruction.
    
    It also creates the MANIFEST.in file if necessary.
    """

    AUTHOR_REGEX = re.compile('(?u)^(?P<name>[- .,\w\d\'’"()]+) <(?P<email>.+?)>$')

    PYTHON_VERSIONS = {
        2: ['2.6', '2.7'],
        3: ['3.0', '3.1', '3.2', '3.3', '3.4', '3.5', '3.6']
    }

    def __init__(self):
        self._manifest = []

    def build(self, poet, **options):
        """
        Builds a package from a Poet instance

        :param poet: The poet to build.
        :type poet: poet.poet.Poet
        """
        setup_kwargs = self._setup(poet, **options)

        setup = os.path.join(poet.base_dir, 'setup.py')
        manifest = os.path.join(poet.base_dir, 'MANIFEST.in')
        self._write_setup(setup_kwargs, setup)

        readme = None
        if poet.has_markdown_readme():
            readme = os.path.join(poet.base_dir, 'README.rst')
            if os.path.exists(readme):
                readme = None
            else:
                self._write_readme(readme, setup_kwargs['long_description'])

        self._manifest.append('include README.rst')

        self._write_manifest(manifest)

        try:
            dist = Distribution(setup_kwargs)
            dist.run_command('sdist')
        except Exception:
            raise
        finally:
            os.unlink(setup)
            os.unlink(manifest)

            if readme:
                os.unlink(readme)

        # Building wheel if necessary
        if not options.get('no_wheels'):
            command = WheelCommand()
            command_args = [
                '--no-index',
                '--no-deps',
                '-q',
                '--wheel-dir', 'dist',
                'dist/{}'.format(poet.archive)
            ]

            if options.get('universal', True):
                command_args.append('--build-option=--universal')

            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=UserWarning)
                status = command.main(command_args)

            if status != SUCCESS:
                raise Exception('An error occurred while executing command.')

    def _setup(self, poet, **options):
        """
        Builds the setup kwargs base on the Poet instance
        
        :param poet: The Poet instance for which to build.
        :type poet: poet.poet.Poet
        
        :rtype: dict
        """
        setup_kwargs = {
            'name': poet.name,
            'version': poet.normalized_version,
            'description': poet.description,
            'long_description': poet.readme,
            'include_package_data': True,
            'script_name': 'setup.py'
        }

        setup_kwargs.update(self._author(poet))

        setup_kwargs['url'] = self._url(poet)

        setup_kwargs['license'] = poet.license

        setup_kwargs['keywords'] = self._keywords(poet)

        setup_kwargs['classifiers'] = self._classifiers(poet)

        setup_kwargs['entry_points'] = self._entry_points(poet)

        setup_kwargs['install_requires'] = self._install_requires(poet)
        setup_kwargs['tests_require'] = self._tests_require(poet)
        setup_kwargs['extras_require'] = self._extras_require(poet)

        setup_kwargs.update(self._packages(poet))

        # Extensions
        setup_kwargs.update(self._ext_modules(poet))

        return setup_kwargs

    def _author(self, poet):
        """
        Build the author information from a Poet instance.
        
        Transforms a author in the form "name <email>" into
        a proper dictionary.
        
        :param poet: The Poet instance for which to build.
        :type poet: poet.poet.Poet
        
        :rtype: dict
        """
        m = self.AUTHOR_REGEX.match(poet.authors[0])

        name = m.group('name')
        email = m.group('email')

        if PY2:
            name = encode(name)
            email = encode(email)

        return {
            'author': name,
            'author_email': email
        }

    def _url(self, poet):
        return poet.homepage or poet.repository

    def _keywords(self, poet):
        return ' '.join(poet.keywords or [])

    def _classifiers(self, poet):
        """
        Builds the classifiers list from the
        specified Python versions.
        
        :param poet: The Poet instance for which to build.
        :type poet: poet.poet.Poet
        
        :rtype: list
        """
        classifers = ['Programming Language :: Python']
        compatible_versions = {}

        for python in poet.python_versions:
            constraint = Spec(python)

            for major in [2, 3]:
                available_versions = self.PYTHON_VERSIONS[major]

                for version in available_versions:
                    if Version.coerce(version) in constraint:
                        if major not in compatible_versions:
                            compatible_versions[major] = []

                        compatible_versions[major].append(version)

        for major in sorted(list(compatible_versions.keys())):
            versions = compatible_versions[major]
            classifer_template = 'Programming Language :: Python :: {}'

            classifers.append(classifer_template.format(major))

            for version in versions:
                classifers.append(classifer_template.format(version))

        return classifers

    def _entry_points(self, poet):
        """
        Builds the entry points
        
        :param poet: The Poet instance for which to build.
        :type poet: poet.poet.Poet
        
        :rtype: list
        """
        entry_points = {
            'console_scripts': []
        }

        # Generic entry points
        for category, entry_points in poet.entry_points.items():
            entry_points[category] = []

            for name, script in entry_points.items():
                entry_points[category].append('{} = {}'.format(name, script))

        # Console scripts entry points
        for name, script in poet.scripts.items():
            entry_points['console_scripts'].append('{} = {}'.format(name, script))

        return entry_points

    def _install_requires(self, poet):
        """
        Builds the dependencies list.
        
        :param poet: The Poet instance for which to build.
        :type poet: poet.poet.Poet
        
        :rtype: dict
        """
        requires = []
        dependencies = poet.dependencies

        for dependency in dependencies:
            if dependency.optional:
                continue

            requires.append(dependency.normalized_name)

        return requires

    def _tests_require(self, poet):
        """
        Builds the dev dependencies list.
        
        :param poet: The Poet instance for which to build.
        :type poet: poet.poet.Poet
        
        :rtype: dict
        """
        requires = []
        dependencies = poet.dev_dependencies

        for dependency in dependencies:
            if dependency.optional:
                continue

            requires.append(dependency.normalized_name)

        return requires

    def _extras_require(self, poet):
        """
        Builds the extras dictionary from
        the configured features.
        
        :param poet: The Poet instance for which to build.
        :type poet: poet.poet.Poet
        
        :rtype: dict
        """
        if not poet.features:
            return {}

        extras = {}
        for feature_name, featured_packages in poet.features.items():
            extras[feature_name] = []

            for package in featured_packages:
                for dep in poet.dependencies:
                    if dep.name == package:
                        extras[feature_name].append(dep.normalized_name)

        return extras

    def _packages(self, poet):
        """
        Builds the packages and modules list
        based on the include and exclude sections.
        
        It will also register files that need to be put
        in the MANIFEST.in file.
        
        :param poet: The Poet instance for which to build.
        :type poet: poet.poet.Poet
        
        :rtype: dict 
        """
        includes = poet.include
        packages = []
        modules = []
        package_dirs = {}
        crawled = []
        excluded = []
        root = Path(poet.base_dir)

        for exclude in poet.exclude + poet.ignore:
            if not exclude:
                continue

            if exclude.startswith('/'):
                exclude = exclude[1:]

            for exc in root.glob(exclude):
                if exc.suffix == '.py':
                    exc = exc.relative_to(root)
                    excluded.append('.'.join(exc.with_suffix('').parts))

        if not isinstance(includes, list):
            includes = [includes]

        for include in includes:
            if isinstance(include, dict):
                settings = self._find_packages_from(
                    root,
                    include['from'],
                    include['include'],
                    include.get('as', ''),
                    excluded=excluded,
                    crawled=crawled
                )
            else:
                settings = self._find_packages_from(
                    root,
                    '',
                    include,
                    excluded=excluded,
                    crawled=crawled
                )

            packages += settings['packages']
            modules += settings['modules']
            package_dirs.update(settings.get('package_dirs', {}))

        packages = [p for p in packages if p not in excluded]
        modules = [m for m in modules if m not in excluded]

        settings = {
            'packages': packages,
            'py_modules': modules
        }

        package_dir = {}
        for package_name, directory in package_dirs.items():
            package_dir[package_name] = directory.as_posix()

        if package_dir:
            settings['package_dir'] = package_dir

        return settings

    def _find_packages_from(self, root, base_dir, includes,
                            package_name=None, excluded=None, crawled=None):
        package_dirs = {}
        packages = []
        modules = []

        if package_name is not None:
            package_dirs[package_name] = Path(base_dir)

        if excluded is None:
            excluded = []

        if crawled is None:
            crawled = []

        if not isinstance(includes, list):
            includes = [includes]

        if not isinstance(base_dir, Path):
            base_dir = Path(base_dir)

        base_path = root / base_dir

        for include in includes:
            dirs = []
            others = []

            for element in base_path.glob(include):
                if element.is_dir():
                    dirs.append(element.relative_to(base_path))
                else:
                    others.append(element.relative_to(base_path))

            m = re.match('^([^./]+)/\*\*/\*(\..+)?$', include)
            if m:
                # {dir}/**/* will not take the root directory
                # So we add it
                dirs.insert(0, Path(m.group(1)))

            for dir in dirs:
                if dir in crawled:
                    continue

                package = '.'.join(dir.parts)

                # We have a package
                real_dir = base_path / dir
                if (real_dir / '__init__.py').exists():
                    children = [
                        c.relative_to(base_path) for c in real_dir.glob('*.py')
                    ]

                    filtered_children = [c for c in children if '.'.join(c.parts) not in excluded]
                    if children == filtered_children:
                        # If none of the children are excluded
                        # We have a full package
                        packages.append(package)
                    else:
                        modules += ['.'.join(c.parts) for c in filtered_children]

                    crawled += [base_path / child for child in children]

                crawled.append(real_dir)

            for element in others:
                if base_path / element in crawled or element.suffix == '.pyc':
                    continue

                if element.suffix == '.py' and element.name != '__init__.py':
                    modules.append('.'.join(element.with_suffix('').parts))
                elif element.suffix not in ['.py', '.pyc'] and '__pycache__' not in element.parts:
                    # Non Python file, add them to data
                    self._manifest.append('include {}\n'.format(element.as_posix()))
                elif element.name == '__init__.py':
                    dir = element.parent
                    real_dir = base_path / dir
                    children = [
                        c.relative_to(base_path)
                        for c in real_dir.glob('*.py')
                        if c.name != '__init__.py'
                    ]

                    if not children and dir not in crawled:
                        # We actually have a package
                        packages.append('.'.join(dir.parts))

                        crawled.append(base_path / dir)

                crawled.append(base_path / element)

        packages = [p for p in packages if p not in excluded]
        modules = [m for m in modules if m not in excluded]

        settings = {
            'packages': packages,
            'modules': modules
        }

        if package_dirs:
            settings['package_dirs'] = package_dirs

        return settings

    def _ext_modules(self, poet):
        """
        Builds the extension modules.
        
        Transforms the extensions section:
        
        [extensions]
        "my.module" = "my/module.c"
        
        to a proper extension:
        
        Extension('my.module', 'my/module.c')
        
        :param poet: The Poet instance for which to build.
        :type poet: poet.poet.Poet
        
        :rtype: dict 
        """
        extensions = []
        for module, source in poet.extensions.items():
            if not isinstance(source, list):
                source = [source]

            extensions.append(Extension(module, source))

        return {
            'ext_modules': extensions
        }

    def _write_setup(self, setup, dest):
        parameters = setup.copy()

        for key in parameters.keys():
            value = parameters[key]

            if value is not None and not isinstance(value, (list, dict)):
                parameters[key] = repr(value)

        setup_template = template('setup.py')
        with open(dest, 'w') as f:
            f.write(setup_template.render(**parameters))

    def _write_manifest(self, manifest):
        with open(manifest, 'w') as f:
            f.writelines(self._manifest)

    def _write_readme(self, readme, content):
        with open(readme, 'w') as f:
            f.write(content)




# -*- coding: utf-8 -*-

import tempfile

import os
import shutil
import subprocess

from packaging.utils import canonicalize_name
from pip.download import unpack_url
from pip.index import Link
from piptools.resolver import Resolver
from piptools.repositories import PyPIRepository
from piptools.scripts.compile import get_pip_command
from piptools.cache import DependencyCache
from piptools.utils import is_pinned_requirement, key_from_req

from .locations import CACHE_DIR
from .package.pip_dependency import PipDependency
from .utils.helpers import call, template


class Installer(object):

    UNSAFE = ['setuptools']

    def __init__(self, command, repository, with_progress=False):
        self._command = command
        self._poet = command.poet
        self._repository = repository
        self._with_progress = with_progress

    def install(self, features=None, dev=True):
        """
        Install packages defined in configuration files.
        
        If a lock file does not exist, it will lock dependencies
        before installing them.
        
        :param features: Features to install
        :type features: list or None
        
        :param dev: Whether to install dev dependencies or not
        :type dev: bool
        
        :rtype: None 
        """
        if not os.path.exists(self._poet.lock_file):
            if features:
                for feature in features:
                    if feature not in self._poet.features:
                        raise ValueError(
                            'Feature [{}] does not exist'
                            .format(feature)
                        )

            self.lock(dev=dev)

            return self.install(features=features, dev=dev)

        lock = self._poet.lock
        if features:
            for feature in features:
                if feature not in lock.features:
                    raise ValueError(
                        'Feature [{}] does not exist'
                        .format(feature)
                    )

        self._command.line('')
        self._command.line('<info>Installing dependencies</>')
        self._command.line('')

        deps = lock.pip_dependencies

        if dev:
            deps += lock.pip_dev_dependencies

        featured_packages = set()
        for feature, packages in lock.features.items():
            if feature in features:
                for package in packages:
                    featured_packages.add(canonicalize_name(package))

        for dep in deps:
            name = dep.name

            # Package is optional but is not featured
            if dep.optional and name not in featured_packages:
                continue

            # Checking Python version
            if dep.is_python_restricted():
                python_version = self._command.python_version
                if not any([python_version in python for python in dep.python]):
                    # If the package is not compatible
                    # with the current Python version
                    # we do not install
                    if self._command.output.is_verbose():
                        self._command.line(
                            ' - Skipping <info>{}</> '
                            '(Specifies Python <comment>{}</> and current Python is <comment>{}</>)'
                            .format(name, ','.join([str(p) for p in dep.python]), python_version)
                        )
                    continue


            cmd = [self._command.pip(), 'install', dep.normalized_name]

            if dep.is_vcs_dependency():
                constraint = dep.pretty_constraint

                # VCS must be updated to be installed
                cmd.append('-U')
            else:
                constraint = dep.constraint.replace('==', '')

            message = (
                ' - Installing <info>{}</> (<comment>{}</>)'
                .format(name, constraint)
            )
            end_message  = (
                'Installed <info>{}</> (<comment>{}</>)'
                .format(name, constraint)
            )
            error_message = 'Error while installing [{}]'.format(name)

            self._progress(cmd, message[3:], end_message, message, error_message)

    def update(self, packages=None, features=None, dev=True):
        if self._poet.is_lock():
            raise Exception('Update is only available with a poetry.toml file.')

        if packages and features:
            raise Exception('Cannot specify packages and features when updating.')

        self._command.line('')
        self._command.line('<info>Updating dependencies</>')
        self._command.line('')

        # Reading current lock
        lock = self._poet.lock
        current_deps = lock.pip_dependencies
        if dev:
            current_deps += lock.pip_dev_dependencies

        # Resolving new dependencies and locking them
        deps = self._poet.pip_dependencies
        if dev:
            deps += self._poet.pip_dev_dependencies

        featured_packages = set()
        for feature, _packages in self._poet.features.items():
            if feature in features:
                for package in _packages:
                    featured_packages.add(canonicalize_name(package))

        # Removing optional packages unless they are featured packages
        deps = [
            dep
            for dep in deps
            if not dep.optional
               or dep.optional and dep.name in featured_packages
        ]

        if packages:
            packages = [p for p in self.resolve(deps) if p['name'] in packages]
        else:
            packages = self.resolve(deps)

        deps = [PipDependency(p['name'], p['version']) for p in packages]

        delete = not packages and not features
        actions = self._resolve_update_actions(deps, current_deps, delete=delete)

        if not actions:
            self._command.line(' - <info>Dependencies already up-to-date!</info>')

            return

        installs = len([a for a in actions if a[0] == 'install'])
        updates = len([a for a in actions if a[0] == 'update'])
        uninstalls = len([a for a in actions if a[0] == 'remove'])

        summary = []
        if updates:
            summary.append('<comment>{}</> updates'.format(updates))

        if installs:
            summary.append('<comment>{}</> installations'.format(installs))

        if uninstalls:
            summary.append('<comment>{}</> uninstallations'.format(uninstalls))

        if len(summary) > 1:
            summary = ', '.join(summary)
        else:
            summary = summary[0]

        self._command.line(' - Summary: {}'.format(summary))

        error = False
        for action, from_, dep in actions:
            cmd = [self._command.pip()]
            description = 'Installing'

            if action == 'remove':
                description = 'Removing'
                cmd += ['uninstall', dep.normalized_name, '-y']
            elif action == 'update':
                description = 'Updating'
                cmd += ['install', dep.normalized_name, '-U']
            else:
                cmd += ['install', dep.normalized_name]

            name = dep.name

            if dep.is_vcs_dependency():
                constraint = dep.pretty_constraint
            else:
                constraint = dep.constraint.replace('==', '')

            version = '<comment>{}</>'.format(constraint)

            if from_:
                if from_.is_vcs_dependency():
                    constraint = from_.pretty_constraint
                else:
                    constraint = from_.constraint.replace('==', '')

                version = '<comment>{}</> -> '.format(constraint) + version

            message = ' - {} <info>{}</> ({})'.format(description, name, version)
            start_message = message[3:]
            end_message = '{} <info>{}</> ({})'.format(description.replace('ing', 'ed'), name, version)
            error_message = 'Error while {} [{}]'.format(description.lower(), name)

            self._progress(cmd, start_message, end_message, message, error_message)

        if not error:
            # If everything went well, we write down the lock file
            features = {}
            for name, featured_packages in self._poet.features.items():
                name = canonicalize_name(name)
                features[name] = [canonicalize_name(p) for p in featured_packages]

            self._write_lock(packages, features)

    def lock(self, dev=True):
        if self._poet.is_lock():
            return

        self._command.line('')
        self._command.line('<info>Locking dependencies to <comment>poetry.lock</></>')
        self._command.line('')

        deps = self._poet.pip_dependencies

        if dev:
            deps += self._poet.pip_dev_dependencies

        packages = self.resolve(deps)
        features = {}
        for name, featured_packages in self._poet.features.items():
            name = canonicalize_name(name)
            features[name] = [canonicalize_name(p) for p in featured_packages]

        self._write_lock(packages, features)

    def resolve(self, deps):
        if not self._with_progress:
            self._command.line(' - <info>Resolving dependencies</>')

            return self._resolve(deps)

        with self._spin(
            '<info>Resolving dependencies</>',
            '<info>Resolving dependencies</>'
        ):
            return self._resolve(deps)

    def _resolve(self, deps):
        # Checking if we should active prereleases
        prereleases = False
        for dep in deps:
            if dep.accepts_prereleases():
                prereleases = True
                break

        constraints = [dep.as_requirement() for dep in deps]

        command = get_pip_command()
        opts, _ = command.parse_args([])

        resolver = Resolver(
            constraints, PyPIRepository(opts, command._build_session(opts)),
            cache=DependencyCache(CACHE_DIR),
            prereleases=prereleases
        )
        matches = resolver.resolve()
        pinned = [m for m in matches if not m.editable and is_pinned_requirement(m)]
        unpinned = [m for m in matches if m.editable or not is_pinned_requirement(m)]
        reversed_dependencies = resolver.reverse_dependencies(matches)

        # Complete reversed dependencies with cache
        cache = resolver.dependency_cache.cache
        for m in unpinned:
            name = key_from_req(m.req)
            if name not in cache:
                continue

            dependencies = cache[name][list(cache[name].keys())[0]]
            for dep in dependencies:
                dep = canonicalize_name(dep)
                if dep not in reversed_dependencies:
                    reversed_dependencies[dep] = set()

                reversed_dependencies[dep].add(canonicalize_name(name))

        hashes = resolver.resolve_hashes(pinned)
        packages = []
        for m in matches:
            name = key_from_req(m.req)
            if name in self.UNSAFE:
                continue

            version = str(m.req.specifier)
            if m in unpinned:
                url, specifier = m.link.url.split('@')
                rev, _ = specifier.split('#')

                version = self._get_vcs_version(url, rev)
                checksum = 'sha1:{}'.format(version['rev'])
            else:
                version = version.replace('==', '')
                checksum = list(hashes[m])

            # Figuring out category and optionality
            category = None
            optional = False

            # Checking if it's a main dependency
            for dep in deps:
                if dep.name == name:
                    category = dep.category
                    optional = dep.optional
                    break

            if not category:
                def _category(child):
                    opt = False
                    cat = None
                    parents = reversed_dependencies.get(child, set())
                    for parent in parents:
                        for dep in deps:
                            if dep.name != parent:
                                continue

                            opt = dep.optional

                            if dep.category == 'main':
                                # Dependency is given by at least one main package
                                # We flag it as main
                                return 'main', opt

                            return 'dev', opt

                        cat, op = _category(parent)

                        if cat is not None:
                            return cat, opt

                    return cat, opt

                category, optional = _category(name)

            # If category is still None at this point
            # The dependency must have come from a VCS
            # dependency. To avoid missing packages
            # we assume "main" category and not optional
            if category is None:
                category = 'main'
                optional = False

            if not isinstance(checksum, list):
                checksum = [checksum]

            # Retrieving Python restriction if any
            python = self._get_pythons_for_package(
                name, reversed_dependencies, deps
            )
            python = list(python)

            if '*' in python:
                # If at least one parent gave a wildcard
                # Then it should be installed for any Python version
                python = ['*']

            package = {
                'name': name,
                'version': version,
                'checksum': checksum,
                'category': category,
                'optional': optional,
                'python': python
            }

            packages.append(package)

        return sorted(packages, key=lambda p: p['name'].lower())

    def _resolve_update_actions(self, deps, current_deps, delete=True):
        """
        Determine actions on depenncies.
        
        :param deps: New dependencies
        :type deps: list[poet.package.PipDependency]
        
        :param current_deps: Current dependencies
        :type current_deps: list[poet.package.PipDependency]
        
        :param delete: Whether to add delete actions or not
        :type delete: bool
        
        :return: List of actions to execute
        :type: list[tuple]
        """
        actions = []
        for dep in deps:
            action = None
            from_ = None
            found = False

            for current_dep in current_deps:
                name = dep.name
                current_name = current_dep.name
                version = dep.normalized_constraint
                current_version = current_dep.normalized_constraint

                if name == current_name:
                    # Existing dependency
                    found = True

                    if version == current_version:
                        break

                    # If version is different we mark it
                    # as to be updated
                    action = 'update'
                    from_ = current_dep
                    break

            if not found:
                # New dependency. We mark it as to be installed.
                action = 'install'

            if action:
                actions.append((action, from_, dep))

        if not delete:
            return actions

        # We need to check if we have to remove
        # any dependency
        for dep in current_deps:
            found = False

            for new_dep in deps:
                if dep.name == new_dep.name:
                    found = True

                    break

            if not found:
                actions.append(('remove', None, dep))

        return actions

    def _get_vcs_version(self, url, rev):
        tmp_dir = tempfile.mkdtemp()
        current_dir = self._poet.base_dir

        try:
            unpack_url(Link(url), tmp_dir, download_dir=tmp_dir, only_download=True)

            os.chdir(tmp_dir)
            call(['git', 'checkout', rev])

            revision = call(['git', 'rev-parse', rev])
            # Getting info
            revision = revision.strip()
            version = {
                'git': url,
                'rev': revision
            }
        except Exception:
            raise
        finally:
            shutil.rmtree(tmp_dir)
            # Going back to current directory
            os.chdir(current_dir)

        return version

    def _write_lock(self, packages, features):
        self._command.line(' - <info>Writing dependencies</>')

        content = self._generate_lock_content(packages, features)

        with open(self._poet.lock_file, 'w') as f:
            f.write(content)

    def _generate_lock_content(self, packages, features):
        lock_template = template('poetry.lock')

        return lock_template.render(
            name=self._poet.name,
            version=self._poet.version,
            packages=packages,
            features=features
        )

    def _get_pythons_for_package(self, name, reversed_dependencies, deps):
        pythons = set()
        if name not in reversed_dependencies:
            # Main dependency
            for dep in deps:
                if name == dep.name:
                    for p in dep.python:
                        pythons.add(str(p))

                    break

            if not len(pythons):
                pythons.add('*')

            return pythons

        parents = reversed_dependencies[name]

        for parent in parents:
            parent_pythons = self._get_pythons_for_package(
                parent, reversed_dependencies, deps
            )

            pythons = pythons.union(parent_pythons)

        if not len(pythons):
            pythons.add('*')

        return pythons

    def _call(self, cmd, error_message):
        try:
            return call(cmd)
        except subprocess.CalledProcessError as e:
            raise Exception(error_message + ' ({})'.format(str(e)))

    def _progress(self, cmd, start_message, end_message, default_message, error_message):
        if not self._with_progress:
            self._command.line(default_message)

            return self._call(cmd, error_message)

        with self._spin(start_message, end_message):
            return self._call(cmd, error_message)

    def _spin(self, start_message, end_message):
        return self._command.spin(start_message, end_message)



# -*- coding: utf-8 -*-

from ...installer import Installer

from .index_command import IndexCommand


class InstallCommand(IndexCommand):
    """
    Install and lock dependencies specified in the <comment>poetry.toml</comment> file.

    install
        { --f|features=* : Features to install. }
        { --no-dev : Do not install dev dependencies. }
        { --no-progress : Do not output download progress. }
    """

    def handle(self):
        features = []
        for feature in self.option('features'):
            if ' ' in feature:
                features += feature.split(' ')
            else:
                features.append(feature)

        dev = not self.option('no-dev')

        installer = Installer(
            self, self._repository,
            with_progress=not self.option('no-progress')
        )

        installer.install(features=features, dev=dev)



# -*- coding: utf-8 -*-

from semantic_version import Spec, SpecItem, Version

from ..version_parser import VersionParser


class Dependency(object):

    def __init__(self, name, constraint, category='main'):
        self._name = name
        self._constraint = constraint
        self._optional = False
        self._accepts_prereleases = False
        self._category = category
        self._python = [Spec('*')]

        if isinstance(constraint, dict):
            if 'python' in constraint:
                python = constraint['python']

                if not isinstance(python, list):
                    python = [python]

                self._python = [Spec(p) for p in python]

            if 'optional' in constraint:
                self._optional = constraint['optional']

            if 'version' in constraint:
                self._constraint = constraint['version']

        self._normalized_constraint = self._normalize(constraint)

    @property
    def name(self):
        return self._name
    
    @property
    def optional(self):
        return self._optional

    @property
    def constraint(self):
        return self._constraint

    @property
    def category(self):
        return self._category

    @property
    def python(self):
        return self._python

    @property
    def pretty_constraint(self):
        constraint = self._constraint

        if self.is_vcs_dependency():
            if 'git' in self._constraint:
                vcs_kind = 'branch'
                if 'rev' in self._constraint:
                    vcs_kind = 'rev'
                    version = self._constraint['rev']
                elif 'tag' in constraint:
                    vcs_kind = 'tag'
                    version = self._constraint['tag']
                else:
                    version = self._constraint.get('branch', 'master')

                constraint = '{} {}'.format(vcs_kind, version)
            else:
                raise ValueError('Unsupport VCS.')

        return constraint

    @property
    def normalized_constraint(self):
        return self._normalized_constraint
    
    @property
    def normalized_name(self):
        normalized_name = self._name
        normalized_constraint = self.normalized_constraint

        if normalized_constraint:
            normalized_name += normalized_constraint

        return normalized_name

    def is_vcs_dependency(self):
        return isinstance(self._constraint, dict) and 'git' in self._constraint

    def accepts_prereleases(self):
        return self._accepts_prereleases

    def is_python_restricted(self):
        return self._python != [Spec('*')]

    def _normalize(self, constraint):
        """
        Normalizes the constraint so that it can be understood
        by the underlying system.
        
        :param constraint: The dependency constraint.
        :type constraint: str or dict
        
        :rtype: str
        """
        if self.is_vcs_dependency():
            # Any VCS dependency is considered prerelease
            self._is_prerelease = True

            return self._normalize_vcs_constraint(constraint)

        version = constraint
        if isinstance(version, dict):
            version = version['version']

        constraint = self._spec(version)
        normalized = []

        for spec in constraint.specs:
            version = spec.spec

            if VersionParser.parse_stability(version) == 'dev':
                self._accepts_prereleases = True

            major, minor, patch, prerelease = (
                version.major, version.minor, version.patch, version.prerelease
            )
            current = '{}.{}.{}'.format(major, minor or 0, patch or 0)
            current = Version(current)

            if spec.kind == SpecItem.KIND_CARET:
                if current.major != 0 or minor is None:
                    upper = current.next_major()
                elif current.minor != 0 or patch is None:
                    upper = current.next_minor()
                else:
                    upper = current.next_patch()

                if prerelease:
                    current = str(current) + '{}'.format(''.join(prerelease))

                normalized.append('>={},<{}'.format(current, upper))
            elif spec.kind == SpecItem.KIND_TILDE:
                if minor is None and patch is None:
                    upper = current.next_major()
                else:
                    upper = current.next_minor()

                upper = '{}.{}.{}'.format(*upper)

                normalized.append('>={},<{}'.format(current, upper))

                if prerelease:
                    current = str(current) + '{}'.format(''.join(prerelease))
            else:
                current = spec.kind + str(current)
                if prerelease:
                    current += '{}'.format(''.join(prerelease))

                normalized.append(current)

        return ','.join(normalized)

    def _normalize_vcs_constraint(self, constraint):
        # Neither setuptools nor distutils support VCS constraint
        # So by default we return nothing
        return ''

    def _spec(self, version):
        try:
            return Spec(version)
        except ValueError:
            return Spec(str(Version.coerce(version)))

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, self.normalized_name)



# -*- coding: utf-8 -*-

import os
import twine.utils

from twine.commands.upload import find_dists, skip_upload
from twine.repository import Repository as BaseRepository
from twine.exceptions import PackageNotFound, RedirectDetected
from twine.package import PackageFile
from requests_toolbelt.multipart import (
    MultipartEncoder, MultipartEncoderMonitor
)


class Repository(BaseRepository):

    def __init__(self, output, repository_url, username, password):
        self._output = output

        super(Repository, self).__init__(repository_url, username, password)

    def register(self, package):
        data = package.metadata_dictionary()
        data.update({
            ":action": "submit",
            "protocol_version": "1",
        })

        self._output.writeln(
            " - Registering <info>{0}</>".format(package.basefilename)
        )

        data_to_send = self._convert_data_to_list_of_tuples(data)
        encoder = MultipartEncoder(data_to_send)
        resp = self.session.post(
            self.url,
            data=encoder,
            allow_redirects=False,
            headers={'Content-Type': encoder.content_type},
        )
        # Bug 28. Try to silence a ResourceWarning by releasing the socket.
        resp.close()

        return resp

    def _upload(self, package):
        data = package.metadata_dictionary()
        data.update({
            # action
            ":action": "file_upload",
            "protocol_version": "1",
        })

        data_to_send = self._convert_data_to_list_of_tuples(data)

        with open(package.filename, "rb") as fp:
            data_to_send.append((
                "content",
                (package.basefilename, fp, "application/octet-stream"),
            ))
            encoder = MultipartEncoder(data_to_send)
            bar = self._output.create_progress_bar(encoder.len)
            bar.set_format(
                " - Uploading <info>{0}</> <comment>%percent%%</>".format(package.basefilename)
            )
            monitor = MultipartEncoderMonitor(
                encoder, lambda monitor: bar.set_progress(monitor.bytes_read / encoder.len)
            )

            resp = self.session.post(
                self.url,
                data=monitor,
                allow_redirects=False,
                headers={'Content-Type': monitor.content_type},
            )

            if resp.ok:
                bar.finish()

                self._output.writeln('')
            else:
                self._output.overwrite('')

        return resp


class Publisher(object):
    """
    Registers and publishes packages to remote repositories.
    """

    def __init__(self, output, repository,
                 username=None, password=None, config_file='~/.pypirc',
                 cert=None, client_cert=None, repository_url=None):
        self._output = output

        config = twine.utils.get_repository_from_config(
            config_file,
            repository,
            repository_url
        )
        config["repository"] = twine.utils.normalize_repository_url(
            config["repository"]
        )

        username = twine.utils.get_username(username, config)
        password = twine.utils.get_password(password, config)
        ca_cert = twine.utils.get_cacert(cert, config)
        client_cert = twine.utils.get_clientcert(client_cert, config)

        self._repository = Repository(
            output, config["repository"], username, password
        )
        self._repository.set_certificate_authority(ca_cert)
        self._repository.set_client_certificate(client_cert)

    def register(self, poet):
        """
        Register a package represented by a Poet instance.
        
        :param poet: The Poet instance representing the package.
        :type poet: poet.poet.Poet
        """
        package = os.path.join(poet.base_dir, 'dist', poet.archive)

        if not os.path.exists(package):
            raise PackageNotFound(
                '"{0}" does not exist on the file system.'.format(package)
            )

        resp = self._repository.register(PackageFile.from_filename(package, None))
        self._repository.close()

        if resp.is_redirect:
            raise RedirectDetected(
                ('"{0}" attempted to redirect to "{1}" during registration.'
                 ' Aborting...').format(self._repository.url,
                                        resp.headers["location"]))

        resp.raise_for_status()

    def upload(self, poet):
        """
        Upload packages represented by a Poet instance.

        :param poet: The Poet instance representing the package.
        :type poet: poet.poet.Poet
        """
        skip_existing = False
        dists = find_dists(
            [
                os.path.join(
                    os.path.join(poet.base_dir, 'dist'),
                    '{}-{}*'.format(poet.name, poet.version)
                )
            ]
        )

        uploads = [i for i in dists if not i.endswith(".asc")]

        for filename in uploads:
            package = PackageFile.from_filename(filename, None)
            skip_message = (
                " - Skipping <comment>{0}</> because it appears to already exist"
                .format(
                    package.basefilename
                )
            )

            # Note: The skip_existing check *needs* to be first, because otherwise
            #       we're going to generate extra HTTP requests against a hardcoded
            #       URL for no reason.
            if skip_existing and self._repository.package_is_uploaded(package):
                self._output.writeln(skip_message)
                continue

            resp = self._repository.upload(package)

            # Bug 92. If we get a redirect we should abort because something seems
            # funky. The behaviour is not well defined and redirects being issued
            # by PyPI should never happen in reality. This should catch malicious
            # redirects as well.
            if resp.is_redirect:
                raise RedirectDetected(
                    ('"{0}" attempted to redirect to "{1}" during upload.'
                     ' Aborting...').format(self._repository.url,
                                            resp.headers["location"]))

            if skip_upload(resp, skip_existing, package):
                self._output.writeln(skip_message)

                continue

            twine.utils.check_status_code(resp)

        # Bug 28. Try to silence a ResourceWarning by clearing the connection
        # pool.
        self._repository.close()



# -*- coding: utf-8 -*-

import sys

try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path

PY2 = sys.version_info[0] == 2
PY3K = sys.version_info[0] >= 3

if PY2:
    unicode = unicode
    basestring = basestring
else:
    unicode = str
    basestring = str


def decode(string, encodings=None):
    if not PY2 and not isinstance(string, bytes):
        return string

    if PY2 and isinstance(string, unicode):
        return string

    if encodings is None:
        encodings = ['utf-8', 'latin1', 'ascii']

    for encoding in encodings:
        try:
            return string.decode(encoding)
        except (UnicodeEncodeError, UnicodeDecodeError):
            pass

    return string.decode(encodings[0], errors='ignore')


def encode(string, encodings=None):
    if not PY2 and isinstance(string, bytes):
        return string

    if PY2 and isinstance(string, str):
        return string

    if encodings is None:
        encodings = ['utf-8', 'latin1', 'ascii']

    for encoding in encodings:
        try:
            return string.encode(encoding)
        except (UnicodeEncodeError, UnicodeDecodeError):
            pass

    return string.encode(encodings[0], errors='ignore')
